#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.tensor` module provides abstractions to work with tensorial
containers and similar.

The main object defined here is :py:class:`~.Tensor`, which is a that acts as
container for any object :py:class:`~.BaseNumerical`, as in
:py:class:`~.TimeSeries`, or :py:class:`~.UniformGridData`. :py:class:`~.Tensor`
behaves as one would expect them to behave: they are high-level interfaces that
support all the mathematical operations. :py:class:`~.Tensor` also inherit the
attributes from their contained object. For example, if the :py:class:`~.Tensor`
contains :py:class:`~.TimeSeries`, the tensor will dynamically acquire all the
methods in such class.

In pratice, the main way to use :py:class:`~.Tensor` objects is through its
derived classes :py:class:`~.Vector` and :py:class:`~.Matrix`, which implement
all those features one might expect from vector calculus.

Consider the following usage example:

.. code-block:: python

   import numpy as np

   from kuibit.timeseries import TimeSeries
   from kuibit.tensor import Vector

   # Fake some data that describe the x, y position of two black holes
   times = np.linspace(0, 2 * np.pi, 100)

   bh1_x = np.sin(times)
   bh1_y = np.cos(times)

   # Not really realistic, but it is fake data
   bh2_x = np.sin(2 * times)
   bh2_y = np.cos(2 * times)

   bh1_centroid = Vector([TimeSeries(times, bh1_x), TimeSeries(times, bh1_y)])
   bh2_centroid = Vector([TimeSeries(times, bh2_x), TimeSeries(times, bh2_y)])

   # If we want to compute vx, vy
   bh1_velocity = bh_centroid.differentiated()  # This is a Vector

   # For the distance with another
   distance = bh1_centroid - bh2_centroid  # This is a Vector

   # The magnitude of the distance
   distance = distance.norm()   # This is a TimeSeries

"""


from __future__ import annotations

from inspect import getsource  # Used in __getattr__
from typing import (
    Any,
    Collection,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

import numpy as np
import numpy.typing as npty

from kuibit.numerical import BaseNumerical

_T = TypeVar("_T", bound=BaseNumerical)


class Tensor(Generic[_T], BaseNumerical):
    """Represents a mathematical hyper-matrix (a Tensor) in N dimensions.

    At the moment, it is only used as a base class for :py:class:`~.Vector` and
    :py:class:`~.Matrix`. So, the class is currently not intended for direct use
    (as it does not have a lot of features). This class is not fully flashed out
    for the general case. It is here mostly as a stub for the two subclasses
    that are below. Implementing a generic tensor class is not trivial and it
    left as an exercise to the reader.

    This class implements the basic infrastrucutre used by :py:class:`~.Vector`
    and :py:class:`~.Matrix`. In particular, it fullfills the requirements set
    by :py:class:`~.BaseNumerical` and it implements a method to broadcast
    attributes from the contained objects to the container itself.

    All the operations are applied component-wise. So, for example, a tensor
    times a tensor is going to be a tensor with the same shape and elements
    multiplied. Reductions return NumPy arrays.

    ..note::

        For efficicency of implmentation, data is sotred without hierarchy
       (i.e., "flattened out"). Therefore, the operations that require
       reconstructing the structure have some small overhead. When, possible
       work with flattened data.


    """

    def __init__(self, data: Collection[_T]):
        """Construct the Tensor by checking and processing the data.

        The data is flattened out and it shape is saved. The shape here is
        defined as the number of elements of the tensor along each dimension.
        For example, for a vector it would only by its length.

        :param data: Representation of the tensor as nested lists.
        :type data: Nested list of data derived from :py:class:`~.BaseNumerical`.

        """
        # We are going to make a design choice here. If we flatten the data, all
        # the operations that are component-wise will become very easy to
        # implement. Also operations that rearrange the indices will become easy
        # The trade-off is that operations that need to know about
        # the actual shape (e.g., maxtrix multiplication) will have to be
        # implemented more carefully.
        if not hasattr(data, "__len__"):
            raise TypeError("data has to be iterable")

        if len(data) == 0:
            raise TypeError("data is empty")

        # Big brain algorithm follows. We want to allow only tensors that have a
        # well defined structure and we want to record this structure. Consider
        # the simple example of a matrix. Matrices look like [[a1, a2], [a3,
        # a4]]. The shape of this matrix would be (2, 2). It does not make sense
        # to have something like [[a1, a2], [a3, [a4]]] (notice the second []
        # around a4). So, below we recursively walk through the data as if it
        # was a tree. We know we reached the lowest level when we find object of
        # type BaseNumerical. The first time we reach the lowest level, we
        # record the depth of that level since all the branches must have the
        # same depth. If any branch has a different depth, it means that the
        # tree is not an hyper-matrix. The second case in which a tree would not
        # be broadcastable as a hyper-matrix is that the number of elements is
        # different for different branches at fixed height, as in this example
        # [[a1, a2], [a3, a4, a5]]. So, at each level we record how many
        # elements we have by putting them in sets. If any set has more than one
        # element at any given time, it means that the data is not like a matrix.

        flattened: List[_T] = []
        shape: Dict[int, Set[int]] = {}

        # Recursively walk through the input, recording the shape
        def _walk(
            data, height: int = 0, depth_of_first_leaf: Optional[int] = None
        ):
            """Walk recursively the tree finding inconsistencies, fattening the
            data, and recording the shape.

            """

            shape.setdefault(height, set()).add(len(data))
            if len(shape[height]) != 1:
                raise RuntimeError("The shape of the data is inconsistent")

            for d in data:
                if isinstance(d, BaseNumerical):
                    # This is a leaf node

                    # Set depth_of_first_leaf to height if it is None
                    depth_of_first_leaf = depth_of_first_leaf or height

                    if height != depth_of_first_leaf:
                        raise RuntimeError("The data has inconsistent depth")

                    flattened.append(d)
                else:
                    depth_of_first_leaf = _walk(
                        d, height + 1, depth_of_first_leaf
                    )

            # The variable `depth_of_first_leaf` has to be carried down all the
            # way, but also "up" from the bottom of the tree. That's why it is
            # the return value, so that we can pass it up when we find it.
            return depth_of_first_leaf

        _walk(data)

        # Change the sets into normal ints
        self.__shape = tuple(v.pop() for v in shape.values())

        self.__flat_data = flattened

        # Check that the type is the same for all the data
        if not all(isinstance(d, self.type) for d in self.flat_data):
            raise TypeError("data has to be of one specific type")

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the shape of the Tensor.

        The shape is defined as the number of elements in each dimension.

        :returns: Shape of the tensor.
        :rtype: Tuple of ints

        """
        return self.__shape

    @classmethod
    def from_shape_and_flat_data(
        cls, shape: Collection[int], flat_data: Collection[_T]
    ) -> Tensor[_T]:
        """Create a new :py:class:`~.Tensor` from flat data and shape.

        No checks are performed.
        """
        # We set the attributes directly, so we don't use __init__ but __new__.
        ret = cls.__new__(cls)
        ret.__shape = tuple(shape)
        ret.__flat_data = list(flat_data)
        return ret

    @property
    def flat_data(self) -> List[_T]:
        """Return the data in the tensor as a flat list.

        :returns: List with all the data
        :rtype: List of type.
        """
        return self.__flat_data

    def _restructure_data(self, flat_data):
        """Take flatten data and reshape it to the correct tensorial form.

        :param flat_data: Data flattened out
        :type: list

        :returns: Data restructured according to self.shape.
        :rtype: Nested list
        """

        # We make groups of groups, recursively, starting from the individual
        # elements.
        #
        # Consider the example with shape (1, 2, 3) and data = [1, 2, 3, 4, 5,
        # 6]. First, we make groups of three: data = [[1, 2, 3], [4, 5, 6]],
        # then we make groups of two: data = [[[1, 2, 3], [4, 5, 6]]], and then
        # we make another group of one. data = [[[[1, 2, 3], [4, 5, 6]]]]. We
        # end up with an extra set of brackets that is due to the fact that we
        # always return a list. So, the return value of this function peels off
        # that last layer.

        def _reconstruct(data, size):
            # Make groups of length `size`
            ret = [data[i : i + size] for i in range(0, len(data), size)]
            return ret

        data = flat_data

        for size in reversed(self.shape):
            data = _reconstruct(data, size)

        # The return value is always a list, so we end up with an extra level of
        # lists, which we remove
        return data[0]

    # TODO: I do not know how to type-hint a return value that is nested lists
    # with an unknown number of levels.
    @property
    def data(self):
        """Return the data with its tensorial structure.

        For example, if this was a Matrix, the return value would be a list of
        lists.

        Note, this operation is not free! Data is stored flat, so recomputing
        the structure has some overhead.

        :returns: Data structured according to self.shape
        :rtype: Nested lists
        """
        return self._restructure_data(self.flat_data)

    @property
    def type(self) -> type:
        """Return the type of the contained object.

        :returns: Type of the object, a class derived from
                  :py:class:`~.BaseNumerical`
        :rtype: type
        """
        return type(self.__flat_data[0])

    def _apply_unary(self, function, *args, **kwargs) -> Tensor[_T]:
        return type(self).from_shape_and_flat_data(
            self.shape,
            [
                # skipcq PYL-W0212
                x._apply_unary(function, *args, **kwargs)
                for x in self.flat_data
            ],
        )

    def _apply_binary(self, other, function, *args, **kwargs) -> Tensor[_T]:
        if isinstance(other, type(self)):
            if self.type != other.type:
                raise ValueError("Incompatible base types")

            if self.shape != other.shape:
                raise ValueError("Tensors do not have same shape")
            else:
                ret = type(self).from_shape_and_flat_data(
                    self.shape,
                    [
                        # skipcq PYL-W0212
                        x._apply_binary(y, function, *args, **kwargs)
                        for x, y in zip(self.flat_data, other.flat_data)
                    ],
                )
        else:
            ret = type(self).from_shape_and_flat_data(
                self.shape,
                [
                    # skipcq PYL-W0212
                    x._apply_binary(other, function, *args, **kwargs)
                    for x in self.flat_data
                ],
            )
        return ret

    def _apply_reduction(self, reduction, *args, **kwargs) -> npty.NDArray:
        return np.array(
            self._restructure_data(
                [
                    # skipcq PYL-W0212
                    x._apply_reduction(reduction, *args, **kwargs)
                    for x in self.flat_data
                ]
            )
        )

    def _apply_to_self(self, function, *args, **kwargs):
        self.__flat_data = [
            function(x, *args, **kwargs) for x in self.flat_data
        ]

    def __call__(self, val) -> npty.NDArray:
        try:
            ret = np.array(
                self._restructure_data([x(val) for x in self.flat_data])
            )
        except AttributeError as e:
            raise AttributeError(f"{self.type} cannot ba called") from e

        return ret

    def copy(self) -> Tensor[_T]:
        """Return a deep copy of the object.

        :return: Deep copy
        :rtype: Tensor
        """
        try:
            ret = type(self).from_shape_and_flat_data(
                self.shape,
                [x.copy() for x in self.flat_data],
            )
        except AttributeError as e:
            raise AttributeError(
                f"{self.type} does not have a copy method"
            ) from e

        return ret

    def __matmul__(self, other):
        return NotImplemented

    def __rmatmul__(self, other):
        return NotImplemented

    def __getattr__(self, attr: str):
        # This is how we transfer all the methods in _T to the Tensor itself.
        #
        # For example, if T is a TimeSeries, we would like to be able to call
        # .partial_differentiated for all the components and create a vector out
        # of it. With this, if V is a Vector of T, we will be able to call
        # V.partial_differentiated and get the desired output.
        #
        # The return value of __getattr__ is a new function, _apply_attr. This
        # _apply_attr is such that it calls the attribute we want to call with
        # the arguments we pass. In addition to that, we do some basic type
        # checking to return the correct type. The only caveat here is that the
        # attribute has to return T or a number.

        # We broadcast only those attributes that are in the base object
        if not hasattr(self.flat_data[0], attr):
            raise AttributeError(f"No attribute {attr} in {self.type}")

        # There are some nasty things that we have to do to ensure nice
        # compatibility: we need to handle @property methods and methods that
        # edit the object in place. For the first, we first need to see if the
        # attribute is a property or not. Property attributes are not callable.

        # Note, we are applying getattr to the class, not to the instance!
        is_property = isinstance(getattr(self.type, attr), property)

        # For the latter, we are going to inspect the source code of the method
        # as see if it calls self._apply_to_self, in which case we will do the
        # same.
        if is_property:
            source_code = getsource(getattr(self.type, attr).fget)
        else:
            source_code = getsource(getattr(self.type, attr))

        applies_to_self = "self._apply_to_self" in source_code

        # If we have a property, we should not return a callable
        if is_property:
            new_data = [getattr(x, attr) for x in self.flat_data]

            if isinstance(new_data[0], self.type):
                return type(self).from_shape_and_flat_data(
                    self.shape, new_data
                )

            if isinstance(new_data[0], (float, complex, np.ndarray)):
                return np.array(new_data).reshape(self.shape)

            raise NotImplementedError(
                "Attribute broadcasting works only for methods that return numbers or type(self)"
            )

        # If we have don't a property, we should return a callable that does
        # pretty much the same thing

        def _apply_attr(*args, **kwargs):
            new_data = [
                getattr(x, attr)(*args, **kwargs) for x in self.flat_data
            ]
            if isinstance(new_data[0], self.type):
                return type(self).from_shape_and_flat_data(
                    self.shape, new_data
                )
            if isinstance(new_data[0], (float, complex, np.ndarray)):
                return np.array(new_data).reshape(self.shape)

            # If the return type is None, this might mean that we are editing
            # something in place. In that case, there's nothing we have to do,
            # as the objects already changed
            if applies_to_self:
                return None

            raise NotImplementedError(
                "Attribute broadcasting works only for methods that return numbers or type(self)"
            )

        return _apply_attr

    def __eq__(self, other: Any) -> bool:
        # Let's take advantage of some shortcircuiting to reduce comparisons. We
        # check three conditions that would establish inequality, and negate the
        # answer. The reason is that if any of them fails, we wouldn't have to
        # go to the next one. This is particularly relevant for the last
        # condition, which is a little expensive to compute.
        not_equal = (
            not isinstance(other, type(self))
            or self.shape != other.shape
            or any(
                s != i
                for s, i in zip(iter(self.flat_data), iter(other.flat_data))
            )
        )
        return not not_equal

    # From Python's docs: In order to conform to the object model, classes that
    # define their own equality method should also define their own hash method,
    # or be unhashable.

    # Since we consider series unhashable, this object also has to be unhashable.
    __hash__ = None


class Vector(Tensor[_T]):
    """Represents a vector in the mathematical sense.

    It can be used with series, or grid data, or anything that is derived from
    :py:class:`~.BaseNumerical`.

    This abstraction is useful for vector operations: for example, taking the
    cross/dot products between two vectors.

    All the operations are component-wise, and the vector inherits all the
    methods available to the base object.

    """

    def __len__(self) -> int:
        """Length of the vector."""
        return len(self.flat_data)

    def __getitem__(self, i: int) -> _T:
        return self.flat_data[i]

    def dot(self, other: Vector[_T]) -> _T:
        """Return the dot product with another vector.

        :param other: Other vector in the dot product.
        :type other: Vector

        :returns: Dot product
        :rtype: Same as self.type
        """
        if not isinstance(other, Vector):
            raise TypeError("other is not a Vector")

        if self.shape != other.shape:
            raise RuntimeError("Incosistent shape")

        if self.type != other.type:
            raise TypeError("Incosistent base type")

        return sum(s * i for s, i in zip(self, other))

    def norm(self):
        """Return the norm of the vector.

        :returns: Norm of the vector
        :rtype: Same as self.type
        """
        return self.dot(self).sqrt()


class Matrix(Tensor[_T]):
    """Represents a matrix in the mathematical sense.

    It can be used with series, or grid data, or anything that is derived from
    :py:class:`~.BaseNumerical`.

    This abstraction is useful for matrix operations: for example, taking the
    determinant.

    All the operations are component-wise, and the matrix inherits all the
    methods available to the base object.

    """

    def __getitem__(self, ij: Tuple[int]) -> _T:
        # Let's see an example, the matrix [[0, 1, 2], [3, 4, 5]] has shape (2,
        # 3) = (X, Y). If we want element (1, 2) = (i, j), we would have to get
        # index i * Y + j
        i, j = ij
        _, Y = self.shape

        return self.flat_data[i * Y + j]



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Based on code originally developed by Wolfgang Kastaun. This file may contain
# algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_scalars.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_scalars` module provides simple interfaces to access
time series data as output by CarpetASCII, including all the reductions.

There are multiple classes defined in this module:

- :py:class`~.ScalarsDir` interfaces with :py:class:`~.SimDir` and organizes the
  data according to their reduction. This is a dictionary-like object with keys
  the possible reduction (e.g., ``max``, ``average``, ``norm2``). -
- :py:class`~.AllScalars` takes all the files that correspond to a given reduction
  and organize them according to the variables they contain.
- :py:class`~.OneScalar` represents one single scalar variable, with data that
  is represented as :py:class:`~.TimeSeries` objects. :py:class`~.AllScalars` contains
  many :py:class`~.OneScalar` objects.

These are hierarchical classes, one containing the others, so one typically ends
up with a series of brackets or dots to access the actual data. For example, if
``sim`` is a :py:class:`~.SimDir`, ``sim.ts.max['rho_b']`` is maximum of
``rho_b`` represented as :py:class:`~.TimeSeries`.

"""

import os
import re
from bz2 import open as bopen
from functools import lru_cache
from gzip import open as gopen

import numpy as np

from kuibit import simdir
from kuibit import timeseries as ts
from kuibit.attr_dict import pythonize_name_dict
from kuibit.cactus_ascii_utils import scan_header


class OneScalar:
    """Read scalar data produced by CarpetASCII.

    :py:class:`~.OneScalar` is a dictionary-like object with keys the variables
    and values the :py:class:`~.TimeSeries`.

    Single variable per file or single file per group are supported. In the
    latter case, the header is inspected to understand the content of the file.
    Compressed files (gz and bz2) are supported too.

    :py:class:`~.OneScalar` represents one scalar file, there can be multiple
    variables inside, (if it was output with ``one_file_per_group = yes``).

    :ivar path: Path of the file.
    :type path: str
    :ivar folder: Path of the folder that contains the file.
    :type folder: str
    :ivar reduction_type: Type of reduction.
    :type reduction_type: str

    """

    # What is this pattern?
    # Let's understand it. We have ^ and $, so we match the entire string and
    # we have seven capturing groups.
    # 1: (\w+) matches any number of characters greater than 0 (w = word)
    # 2: ((-(\w+))|(\[\d+\]))? optionally match one of the two
    # 3: Matches - with followed by 4: any word
    # 5: Matches brackets with a number inside
    # In between match a dot (\.)
    # 6: (minimum|maximum|norm1|norm2|norm_inf|average|scalars)? optionally match one
    #    of those
    # In between match .asc (\.asc)
    # 7: (\.(gz|bz2))? optionally match .gz or .bz2

    # We want to match file names like hydrobase-press.maximum.asc or
    # hydrobase-vel[0].maximum.asc
    #
    # The .scalars. file is the one generated with the option
    # all_reductions_in_one_file

    _pattern_filename = r"""
    ^(\w+)
    ((-(\w+))|(\[\d+\]))?
    \.(minimum|maximum|norm1|norm2|norm_inf|average|scalars)?
    \.asc
    (\.(gz|bz2))?$"""

    _rx_filename = re.compile(_pattern_filename, re.VERBOSE)

    _reduction_types = {
        "minimum": "min",
        "maximum": "max",
        "norm1": "norm1",
        "norm2": "norm2",
        "norm_inf": "infnorm",
        "average": "average",
        None: "scalar",
    }

    # What function to use to open the file?
    # What mode?
    _decompressor = {
        None: (open, "r"),
        "gz": (gopen, "rt"),
        "bz2": (bopen, "rt"),
    }

    def __init__(self, path):
        """Constructor.

        Here we understand what the file contains.

        :param path: Path of the file.
        :type path: str
        """
        self.path = str(path)
        # The _vars_columns dictionary contains a mapping between the various variables
        # and the column numbers in which they are stored.
        self._vars_columns = {}

        # The _vars dictionary maps variables with the TimeSeries. It is used to
        # cached loaded data.
        self._vars = {}

        self.folder, filename = os.path.split(self.path)

        filename_match = self._rx_filename.match(filename)

        if filename_match is None:
            raise RuntimeError(f"Name scheme not recognized for {filename}")

        # variable_name1 may be a thorn name (e.g. hydrobase-press)
        # Or the actual variable name (e.g. H)
        (
            variable_name1,
            _0,
            _1,
            variable_name2,
            index_in_brackets,
            reduction_type,
            _2,
            compression_method,
        ) = filename_match.groups()

        self._compression_method = compression_method

        self.reduction_type = (
            reduction_type if reduction_type is not None else "scalar"
        )

        # If the file contains multiple variables, we will scan the header
        # immediately to understand the content. If not, we scan the header
        # only when needed
        self._is_one_file_per_group = variable_name2 is not None
        self._was_header_scanned = False

        if self._is_one_file_per_group:
            # We need the variable names
            self._scan_header()
        else:
            variable_name = variable_name1
            if index_in_brackets is not None:
                variable_name += index_in_brackets
            self._vars_columns = {variable_name: None}

    def _scan_header(self):
        # Call scan_header with the right argument

        extended_format = self.reduction_type == "scalar"

        # What method to we need to use to open the file?
        # opener can be open, gopen, or bopen depending on the extension
        # of the file
        opener, opener_mode = self._decompressor[self._compression_method]

        self._time_column, columns_info = scan_header(
            self.path,
            self._is_one_file_per_group,
            extended_format,
            opener=opener,
            opener_mode=opener_mode,
        )

        if self._is_one_file_per_group:
            self._vars_columns.update(columns_info)
        else:
            # There is only one data_column
            self._vars_columns = {
                list(self._vars_columns.keys())[0]: columns_info
            }

        self._was_header_scanned = True

    @lru_cache(128)
    def load(self, variable):
        """Read file and return a TimeSeries with the requested variable.

        :param variable: Requested variable.
        :type variable: str

        :returns: :py:class:`~.TimeSeries` with requested variable as read from
                  file
        :rtype: :py:class:`~.TimeSeries`

        """
        if not self._was_header_scanned:
            self._scan_header()

        if variable not in self:
            raise KeyError(f"{variable} not available")

        column_number = self._vars_columns[variable]
        t, y = np.loadtxt(
            self.path,
            unpack=True,
            ndmin=2,
            usecols=(self._time_column, column_number),
        )

        return ts.remove_duplicated_iters(t, y)

    def __getitem__(self, key):
        if key not in self._vars:
            # Error checking is done in __load__
            self._vars[key] = self.load(key)
        return self._vars[key]

    def __contains__(self, key):
        return key in self._vars_columns

    def keys(self):
        """Return the list of variables available.

        :returns: Variables in the file
        :rtype:   dict_keys

        """
        return self._vars_columns.keys()


class AllScalars:
    """Helper class to read various types of scalar data in a list of files and
    properly order them. The core of this object is the ``_vars`` dictionary
    which contains the location of all the files for a specific variable and
    reduction (as :py:class:`~.OneScalar`).

    :py:class:`~.AllScalars` is a dictionary-like object, using the bracket notation
    you can access values with as TimeSeries. Alternatively, you can access the
    data as attributes of the ``fields`` attribute.

    :ivar reduction_type: Type of reduction.
    :type reduction_type: str

    """

    def __init__(self, allfiles, reduction_type):
        """Constructor.

        :param allfiles: List of all the files
        :type allfiles: list of str
        :param reduction_type: Type of reduction.
        :type reduction_type: str

        """
        self.reduction_type = str(reduction_type)

        # TODO: Is it necessary to have the folder level?
        # Probably not, so remove it

        # _vars_readers is like _vars_readers['variable']['folder'] ->
        # OneScalar(f) _vars_readers['variable'] is a dictionary with as keys
        # the folders where to find the files associated to the variable and the
        # reduction reduction_type
        #
        # OneScalar objects act as readers.
        self._vars_readers = {}
        for file_ in allfiles:
            # We only save those that variables are well-behaved
            try:
                cactusascii_file = OneScalar(file_)
                if cactusascii_file.reduction_type == reduction_type:
                    for var in list(cactusascii_file.keys()):
                        # We add to the _vars_readers dictionary the mapping:
                        # [var][folder] to OneScalar(f)
                        folder = cactusascii_file.folder
                        self._vars_readers.setdefault(var, {})[
                            folder
                        ] = cactusascii_file
            except RuntimeError:
                pass

        # We cache the results in _vars, a dictionary with keys the variables
        # and values the timeseries.
        self._vars = {}

        # What pythonize_name_dict does is to make the various variables
        # accessible as attributes, e.g. self.fields.rho
        self.fields = pythonize_name_dict(list(self.keys()), self.__getitem__)

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(f"{key} not available")

        if key not in self._vars:
            # We read all the files associated to variable key
            folders = self._vars_readers[key]
            series = [f[key] for f in folders.values()]
            self._vars[key] = ts.combine_ts(series)

        return self._vars[key]

    def __contains__(self, key):
        return key in self._vars_readers

    def keys(self):
        """Return the available variables corresponding to the given reduction.

        :returns: Variables with given reduction
        :rtype:   dict_keys

        """
        return self._vars_readers.keys()

    def get(self, key, default=None):
        """Return variable if available, else return the default value.

        :param key: Requested variable.
        :type key: str
        :param default: Returned value if ``variable`` is not available.
        :type default: any

        :returns: :py:class:`~.TimeSeries` of the requested variable
        :rtype: :py:class:`~.TimeSeries`

        """
        if key in self:
            return self[key]

        return default

    def __str__(self):
        ret = f"Available {self.reduction_type} timeseries:\n"
        ret += f"{list(self.keys())}\n"
        return ret


class ScalarsDir:
    """This class provides acces to various types of scalar data in a given
    simulation directory. Typically used from a :py:class:`~.SimDir` instance.
    The different scalars are available as attributes:

    :ivar scalar:    access to grid scalars.
    :ivar minimum:   access to minimum reduction.
    :ivar maximum:   access to maximum reduction.
    :ivar norm1:     access to norm1 reduction.
    :ivar norm2:     access to norm2 reduction.
    :ivar average:   access to average reduction.
    :ivar infnorm:   access to inf-norm reduction.

    Each of those works as a dictionary mapping variable names to
    :py:class:`~.TimeSeries` instances.

    """

    # TODO: Implement the following, possibly in a clean way

    # .. note::
    #    infnorm is reconstructed from min and max if infnorm
    #    itself is not available.

    def __init__(self, sd):
        """The constructor is not intended for direct use.

        :param sd: Simulation directory
        :type sd:  :py:class:`~.SimDir` instance.
        """
        if not isinstance(sd, simdir.SimDir):
            raise TypeError("Input is not SimDir")

        self.path = sd.path
        self.point = AllScalars(sd.allfiles, "scalar")
        self.scalar = AllScalars(sd.allfiles, "scalar")
        self.minimum = AllScalars(sd.allfiles, "minimum")
        self.maximum = AllScalars(sd.allfiles, "maximum")
        self.norm1 = AllScalars(sd.allfiles, "norm1")
        self.norm2 = AllScalars(sd.allfiles, "norm2")
        self.average = AllScalars(sd.allfiles, "average")
        self.infnorm = AllScalars(sd.allfiles, "infnorm")

        # Aliases
        self.max = self.maximum
        self.min = self.minimum

    def __getitem__(self, reduction):
        return getattr(self, reduction)

    def get(self, key, default=None):
        """Return a reduction if available, else return the default value.

        :param key: Requested reduction.
        :type key: str
        :param default: Returned value if ``reduction`` is not available.
        :type default: any

        :returns: Collection of all the variables with a given reduction.
        :rtype: :py:class:`~.AllScalars`

        """
        if key in [
            "point",
            "scalar",
            "minimum",
            "maximum",
            "norm1",
            "norm2",
            "average",
            "infnorm",
        ]:
            return self[key]

        return default

    def __str__(self):
        str_ = [f"Folder {self.path}"]
        for key in (
            self.path,
            self.scalar,
            self.minimum,
            self.maximum,
            self.norm1,
            self.norm2,
            self.average,
        ):
            str_.append(str(key))
        return "\n".join(str_)



#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_twopunctures` module reads the metadata output by TwoPunctures.

The module is rather straightforward and has only one dictionary-like object
that reads the ``TwoPunctures.bbh`` file created by ``TwoPunctures``.

"""
from __future__ import annotations

import configparser
from collections.abc import KeysView
from typing import Union


class TwoPuncturesDir:
    """Read and process metadata from TwoPunctures.

    This object represents the metadata read from `TwoPunctures.bbh` as a
    dictionary.
    """

    def __init__(self, sd) -> None:
        """Read and process the `TwoPunctures.bbh` file, if in the SimDir."""

        # This is where we store the output
        self._metadata = None

        _metadata_files = [
            path for path in sd.allfiles if path.endswith("TwoPunctures.bbh")
        ]

        if _metadata_files:
            if len(_metadata_files) > 1:
                raise RuntimeError(
                    f"Multiple TwoPunctures.bbh found: {_metadata_files}"
                )

            # We have only one TwoPunctures.bbh
            #
            # The structure of the file is essentially a .ini file, so we can
            # read it with configparser
            data = configparser.ConfigParser()
            data.read(*_metadata_files)

            def try_convert(what: Union[str, float]) -> Union[str, float]:
                # Convert what to a float if it makes sense
                try:
                    ret = float(what)
                except ValueError:
                    ret = what
                return ret

            self._metadata = {
                k: try_convert(v) for k, v in data["metadata"].items()
            }

    @property
    def has_metadata(self) -> bool:
        """Was a TwoPunctures.bbh file found and read?"""
        return self._metadata is not None

    def __getitem__(self, key: str) -> Union[str, float]:
        if self.has_metadata:
            # Everything is stored lowercase
            key_l = key.lower()
            if key_l in self._metadata:
                return self._metadata[key_l]
            raise KeyError(f"{key} is not a metadata for TwoPunctures")
        raise RuntimeError("Metadata not available")

    def keys(self) -> KeysView:
        if self.has_metadata:
            return self._metadata.keys()
        return {}.keys()



#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_timertree.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_timers` module provides an interface to Cactus timers.

Currently, the only timers supported are from the XML output by Carpet with the
option ``output_xml_timer_tree = yes``.

The main class defined in this module is :py:class:`~.TimersDir`. This is a
dictionary-like object that contains most of the information available in the
timers. The keys of this class are the various MPI ranks (ie, process numbers)
found in the output. The value are :py:class:`~.Tree` objects. These trees
contain the reported call-stack of the simulation with the associated timing
information. Simulation restarts are summed up. Often, it is useful to know how
much a function took on average across different MPI ranks. This can be accessed
with the :py:meth:`TimerDir.average` method.

Example of usage:

.. code-block:: python

   # Assuming sim is a SimDir
   avg_timers = sim.timers.average

   # avg_timers.name contains the name of the top level function, main
   # avg_timers.value contains the time took on average by such function

   # avg_timers.children contains a tuple with all the functions called by main.
   # The type of these functions is Tree, so they have a name and value attribute.

   # By default, the units are second. If you want to normalize to 1 and get the
   # percentual value, you can use
   normalized_avg_timers = avg_timers / avg_timers.tot_value_leaves

   print(normalized_avg_timers.value)   # Should be 1

   # Another useful feature is to represent the timer tree with a dictionary or a
   # json file, for postprocessing
   print(normalized_avg_timers.to_json())

"""

import os
import re
from collections.abc import KeysView
from statistics import mean, median
from typing import TYPE_CHECKING, Dict, List, Optional
from xml.etree.ElementTree import Element  # skipcq: BAN-B405
from xml.etree.ElementTree import parse as parse_xml  # skipcq: BAN-B405

from kuibit.tree import Tree, merge_trees

# We need this here to avoid circular imports (because SimDir imports
# cactus_timers)
if TYPE_CHECKING:  # pragma: no cover
    from kuibit.simdir import SimDir


class TimersDir:
    """Timer information read from Carpet.

    This is a dictionary-like object that has as keys the process number and as
    values the timer information represented as a :py:class:`~.Tree` object.

    Timers are summed up across different restarts.

    The most useful method in this class is :py:meth:`~.average`, which returns
    the average timing information across all the MPI ranks.

    Files are read only when needed. At the moment, only XML timers are
    supported, but different backends can be added.

    """

    @staticmethod
    def _load_xml(path: str) -> Tree:
        """Load an xml file to a Tree.

        The structure of the xml has to be more or less like:

        ..code-block:xml

          <timer name = "level1"> 1
            <timer name = "level2a"> 2 </timer>
            <timer name = "level2b"> 3.0 </timer>
            <timer name = "level2c"> 4.0
              <timer name = "level3"> 5 </timer>
            </timer>
          </timer>

        There can be more information, but it will be ignored. We only extract
        the timer tags, read the names and the associated values. When a value
        is not set, we assume it is zero.

        """

        def _traverse_recursive(node: Element) -> Tree:
            """Recursively go through the tree and transform XML to Tree.

            Take the current node, read the name, value, then apply this
            function recursively to the sub-nodes.
            """
            name = node.attrib["name"]
            # If the node does not have a time information attached to it, we just
            # add zero
            value = float(node.text.strip()) if node.text else 0.0
            children_nodes = node.findall("timer")
            children = tuple(map(_traverse_recursive, children_nodes))

            return Tree(name, value, children)

        root_node = parse_xml(path).getroot()  # skipcq: BAN-B314
        return _traverse_recursive(root_node)

    def __init__(self, sd: "SimDir"):
        """Constructor."""
        # Find all the timertree.PROC.xml files and organize them by process number

        # self.tree_files is a dictionary that has as keys the process number
        # and as values list containing all the files associated to that process
        self.tree_files: Dict[int, List[str]] = {}

        # self.timertrees maps process number with a Tree for that process
        # number
        self.timertrees: Dict[int, Tree] = {}

        # We match files with name timertree.NUMBER.xml
        # NUMBER is the process number
        timertree_xml_pattern = r"^timertree.([\d]+).xml$"

        timertree_xml_rx = re.compile(timertree_xml_pattern)

        for path in sd.allfiles:
            _folder, filename = os.path.split(path)
            filename_match = timertree_xml_rx.match(filename)
            if filename_match is not None:
                process_number = int(filename_match.group(1))
                self.tree_files.setdefault(process_number, []).append(path)

        # Let's sort the dictionary by key
        self.tree_files = dict(sorted(self.tree_files.items()))

        self.__average: Optional[Tree] = None
        self.__median: Optional[Tree] = None

    def __getitem__(self, process_number: int) -> Tree:
        """Read total timer information for a given process number.

        :param process_number: MPI rank.
        """
        if process_number not in self.timertrees:
            self.timertrees[process_number] = merge_trees(
                [
                    self._load_xml(path)
                    for path in self.tree_files[process_number]
                ]
            )

        return self.timertrees[process_number]

    def keys(self) -> KeysView:
        """Return the process numbers for which timer information is available."""
        # Process numbers available to read
        return self.tree_files.keys()

    @property
    def average(self) -> Tree:
        """Return the average timer information across all the processes."""
        if self.__average is None:
            # We call self[process_number] explicitely to ensure that everything
            # is loaded
            self.__average = merge_trees(
                [self[process_number] for process_number in self.keys()], mean
            )

        return self.__average

    @property
    def median(self) -> Tree:
        """Return the median timer information across all the processes."""
        if self.__median is None:
            # We call self[process_number] explicitely to ensure that everything
            # is loaded
            self.__median = merge_trees(
                [self[process_number] for process_number in self.keys()],
                median,
            )

        return self.__median

    def __str__(self) -> str:
        return f"Timers available for processes {list(self.keys())}"



#!/usr/bin/env python3

# Copyright (C) 2022-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.tree` module provides a data structure to work with trees.

The :py:class:`~.Tree` describes one node of a tree with a name and a value. The
node has any number of children, which are objects of the type
:py:class:`~.Tree` themselves. (There is no real distinction on what is a node
and what is a tree.) Given a :py:class:`~.Tree`, the children can be accessed by
index or by name. :py:class:`~.Tree` can also be exported to dictionaries or
JSON strings.

Different trees can be combined with the function :py:func:`~.merge_trees`,
which acts in the followin way. It takes any number of nodes that share the name
and create a new tree with that name. The children of this new tree are obtained
by combining all the children of the given nodes, and when children have the
same name, the new value is obtained by applying a given function (e.g., summing
up).

Currently, the main use of this structure in ``kuibit`` is for timers.

"""

# from __future__ import annotations allows us to postpone the evaluation of
# type hints. Without it, we would not be able to define type hints for a
# recursive structure.
from __future__ import annotations

import json
from typing import Callable, Iterable, Union


class Tree:
    """Represent one node of a tree (and recursively, the tree itself).

    :ivar name: Name of the node.
    :ivar value: Value of the node.
    :ivar children: Tuple with the nodes that are children of this one.
    """

    def __init__(
        self, name: str, value: float, children: tuple[Tree, ...] = ()
    ):
        """Constructor."""
        self.name = name
        self.value = value
        self.children = children

    @property
    def is_leaf(self) -> bool:
        """Return whether the node is a leaf node or not."""
        return len(self.children) == 0

    @property
    def tot_value_leaves(self) -> float:
        """Return the sum of all the values in all the levas.

        If the node has no children, return the value of the node itself.

        :returns: Cumulative values of the children.
        """
        if not self.is_leaf:
            return sum(child.tot_value_leaves for child in self.children)

        # We have no children
        return self.value

    def __getitem__(self, child_num_or_name: Union[int, str]) -> Tree:
        if isinstance(child_num_or_name, int):
            try:
                return self.children[child_num_or_name]
            except IndexError:
                raise KeyError(
                    f"Child number {child_num_or_name} not available"
                )
        else:
            for child in self.children:
                if child.name == child_num_or_name:
                    return child

            raise KeyError(f"Child named {child_num_or_name} not available")

    def to_dict(self) -> dict:
        """Convert the tree into a dictionary.

        The conversion happens in the following way: each node is converted into
        a dictionary with two or three elements: the name, the value, and, if
        there are children, a list of children. In turn, the children are
        represented as dictionaries in the same way.

        :returns: A dictionary representing the node and all its children.
        """

        ret_dict = {"name": self.name, "value": self.value}
        if not self.is_leaf:
            ret_dict["children"] = tuple(
                child.to_dict() for child in self.children
            )

        return ret_dict

    def to_json(self) -> str:
        """Convert the tree into a string with its JSON representation.

        :returns: String containing the JSON representation of the tree.
        """
        return json.dumps(self.to_dict())

    def __truediv__(self, value: float) -> Tree:
        """Return a new tree with all the values are divided by the given one."""

        # TODO (REFACTORING): Implement other arithmetic operations
        #
        # It would be straightforward to abstract away this method and implement
        # all the arithmetic operations. We just need to call the correct method
        # when setting value=...
        if not isinstance(value, (int, float)):
            raise TypeError("Given value is not a number")

        def walk_rec(node: Tree) -> Tree:
            return type(self)(
                name=node.name,
                value=node.value / value,
                children=tuple(map(walk_rec, node.children)),
            )

        return walk_rec(self)

    def __str__(self) -> str:
        return f"{self.name}: {self.value}"

    def __eq__(self, other: Tree) -> bool:
        return (
            self.name == other.name
            and self.value == other.value
            and self.children == other.children
        )

    def __hash__(self) -> int:
        """Return the hash of this object.

        Since the entire tree has to be traversed, this is an expensive
        operation!

        """
        # We add the hash mostly because we also have __eq__. Realistically,
        # this function should not be used. We combine hashes by taking the hash
        # of the tuple.
        return hash((self.name, self.value, self.children))


def merge_trees(
    trees: Iterable[Tree],
    merge_function: Callable[[Iterable[float]], float] = sum,
) -> Tree:
    """Combine multiple trees that start from the same place to a new one.

    When multiple nodes are found at the same level with the same name, apply
    `merge_function` to the list of values to generate the new value. The
    default is to sum them up, but another good idea would be to take the mean.

    The algorithm that mergers the tree is simple: it combines all the children
    of any given node (as identified by the name) across the tree. Therefore,
    the trees are meaningfully merged only if they are already relatively
    similar one with the other.

    :param trees: List of trees that have to be merged. They have to start from
                  a node with the same name.

    :param merge_function: Function that has to be applied to reduce the various
                           values to a single one.

    :returns: A new tree with all the nodes from the given trees.

    """

    def walk_rec(nodes: Iterable[Tree]) -> Tree:
        # Check if all the trees start from the same node. We put the names into a
        # set, if the size of the set is not 1, then it means we have different
        # initial nodes.
        root_names = {node.name for node in nodes}
        if len(root_names) != 1:
            raise RuntimeError("Trees do not start with the same base node")

        # This is the algorithm we implement to merge different trees. The
        # tricky part here is to extend the children list or aggregate children
        # with the same name using the function `merge_function`.
        #
        # First, we combine all the children in the same list. This is going to
        # include duplicates, i.e. children with the same name. The next step is
        # to apply this function recursively on each subgroup of children that
        # share the same name. In the process, we collect the Trees as new
        # children. Eventually, we reach the case in which we do not have any
        # children. Threfore, the list comprehension that activates the
        # recursion is not invoked (because child_groups is empty), so recursion
        # stops.

        # All the children including duplicates
        children: list[Tree] = []
        for node in nodes:
            children.extend(node.children)

        # Form the subgroups that share the same name. children_groups is a
        # dictionary that maps names with lists of Trees (having the same name)
        children_groups: dict[str, list[Tree]] = {}
        # We go over all the children and sort them
        for child in children:
            children_groups.setdefault(child.name, []).append(child)

        # Now, we apply recursively this function on the various subgroups
        pruned_children = [
            walk_rec(nodes) for nodes in children_groups.values()
        ]

        # root_names is a set, root_names.pop() is the only element
        return Tree(
            name=root_names.pop(),
            value=merge_function(node.value for node in nodes),
            children=tuple(pruned_children),
        )

    return walk_rec(trees)



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_ah.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_horizons` module provides classes to access the
information about apparent horizons from QuasiLocalMeasures and AHFinderDirect
(including shape files).

The main class is :py:class:`~.HorizonsDir` which collects all available data
from a :py:class:`~.SimDir`. This is a dictionary-like object whose values can
be accessed providing two integers. The first is the QuasiLocalMeasures index,
the second is the AHFinderDirect (we need two indices because at the moment we
have no way to connect the two indexing systems). If only one of the two is
available, the other index can be a dummy.

Once an horizon is selected with a combination of QLM and AHFinderDirect
indices, the resulting object is a :py:class:`~.OneHorizon`.
:py:class:`~.OneHorizon` contains as attributes all the QLM variables.
Similarly, the attribute ``ah`` contains all the variables read from
AHFinderDirect. All of these are represented as :py:class:`~.TimeSeries`.
:py:class:`~.OneHorizon` also contains the shapes of the horizons (if the files
are found), which can be accessed with the methods
:py:meth:`~.shape_at_iteration` and :py:meth:`~.shape_outline_at_iteration`. If
VTK data is available (in ``.vtk`` files), it is processed and stored in the
``vtk`` attribute. :py:func:~.vtk_at_it` is a function that takes an iteration
and returns a dictionary that contains the variables defined on the horizon
(seen as a 3D mesh).
"""

import os
import re
import warnings
from typing import Dict, List, Optional

import numpy as np
from numpy.typing import NDArray

from kuibit.attr_dict import AttributeDictionary, pythonize_name_dict
from kuibit.series import sample_common
from kuibit.timeseries import TimeSeries, combine_ts, remove_duplicated_iters


def compute_horizons_separation(horizon1, horizon2, resample=True):
    """Compute the coordinate separation between the centroids of two horizons.

    The information from the apparent horizons is used (contained in the
    BHDiagnostics files).

    Deprecated in favor of :py:func:`~.compute_separation`.

    :param horizon1: First horizon.
    :type horizon1: :py:class:`~.OneHorizon`
    :param horizon2: Second horizon.
    :type horizon2: :py:class:`~.OneHorizon`

    :returns: Coordinate distance between the two centroids, sampled over both
              the horizons are available.
    :rtype: :py:class:`~.TimeSeries`

    """
    warnings.warn(
        "cactus_horizons.compute_horizons_separation was moved to "
        "hor_utils.compute_separation "
        "and it will be removed in kuibit 1.6.0",
        category=FutureWarning,
    )

    # We add sample_common to make sure that everything is defined on the same
    # interval.
    (
        cen1_x,
        cen1_y,
        cen1_z,
        cen2_x,
        cen2_y,
        cen2_z,
    ) = sample_common(
        (
            horizon1.ah.centroid_x,
            horizon1.ah.centroid_y,
            horizon1.ah.centroid_z,
            horizon2.ah.centroid_x,
            horizon2.ah.centroid_y,
            horizon2.ah.centroid_z,
        ),
        resample=resample,
    )

    return np.sqrt(
        (cen1_x - cen2_x) ** 2
        + (cen1_y - cen2_y) ** 2
        + (cen1_z - cen2_z) ** 2
    )


class OneHorizon:
    r"""This class represents properties of an apparent horizon
    computed from the quasi-isolated horizon formalism.

    All the variables from QuasiLocalMeasures are available as
    :py:class:`~.TimeSeries` as attributes. All the ones from AHFinderDirect are
    attributes of the attribute ``ah``.

    :ivar formation_time: First time at which the horizon has been found
                          (as from AHFinderDirect).
    :ivar mass_final: QLM mass at the last iteration available.
    :ivar spin_final: QLM spin (angular momentum) at the last iteration available.
    :ivar dimensionless_spin_final: Dimensionless spin computed from QLM variables
                                    at the last iteration available.
    :ivar shape_available: Whether the shape files are available.
    :ivar shape_iterations: Iterations at which the shape is available.
    :ivar shape_iterations_min: First iteration at which the shape is available.
    :ivar shape_iterations_max: Last iteration at which the shape is available.
    :ivar shape_times: Times at which the shape is available.
    :ivar shape_times_min: First time at which the shape is available.
    :ivar shape_times_max: Last time at which the shape is available.
    :ivar vtk_available: Whether VTK files are available.
    :ivar ah: All the variables from AHFinderDirect as :py:class:`~.TimeSeries`.

    """

    def __init__(
        self,
        qlm_vars: Optional[Dict[str, TimeSeries]],
        ah_vars: Optional[Dict[str, TimeSeries]],
        shape_files: Optional[Dict[int, str]],
        vtk_files: Optional[Dict[int, str]],
    ):
        """Constructor.

        :param qlm_vars: Dictionary that maps the name of the QLM variable with
                         the associated :py:class:`~.TimeSeries`.
        :type qlm_vars: dict
        :param ah_vars: Dictionary that maps the name of the AH variable with
                        the associated :py:class:`~.TimeSeries`.
        :type ah_vars: dict
        :param shape_files: Dictionary that maps the iteration to the files where
                            to find the shape at that iteration.
        :type shape_files: dict
        :param vtk_files: Dictionary that maps the iteration to the files where
                          to find the vtk information at that iteration.
        :type vtk_files: dict

        """

        # `self.XXXX = XXXX or {}` means set self.XXXX to XXXX if it is not
        # None, otherwise set it to the empty dictionary.

        self._qlm_vars = qlm_vars or {}
        self._ah_vars = ah_vars or {}

        self.ah_available = self._ah_vars != {}
        self.qlm_available = self._qlm_vars != {}

        # We turn the var_dictionary into attributes
        for var, timeseries in self._qlm_vars.items():
            # With this we can access properties in the following way:
            # horizon.mass
            setattr(self, var, timeseries)

        # Here we compute some interesting and useful quantities, if we have
        # qlm data
        if self._qlm_vars:
            self.mass_final = self.mass.y[-1]
            self.spin_final = self.spin.y[-1]
            self.dimensionless_spin_final = (
                self.spin_final / self.mass_final**2
            )
        else:
            self.mass_final = None
            self.spin_final = None
            self.dimensionless_spin_final = None

        # We put the AH vars under the ah attribute, they are accessed in the
        # same way as qlm_vars (as attribute). This is achieved using
        # pythonize_name_dict. The second argument is how we access the data.
        # Here we use the method get_ah_property that peeks into self._ah_vars.
        self.ah = pythonize_name_dict(self._ah_vars, self.get_ah_property)

        # We read the formation time from a variable in AH
        if self._ah_vars:
            self.formation_time = self.ah.area.tmin
        else:
            self.formation_time = None

        # Alias
        self.time_found = self.formation_time

        # Now we deal with the shape. Shape files is a dictionary that maps
        # iteration to the associated file
        self._shape_files = shape_files or {}

        self.shape_available = self._shape_files != {}

        if self.shape_available:
            # We sort the iterations
            self.shape_iterations = np.array(
                sorted(s for s in self._shape_files)
            )

            self.shape_iteration_min = self.shape_iterations[0]
            self.shape_iteration_max = self.shape_iterations[-1]

            # To convert between time and iteration, we need the AH data If we
            # don't have that, we will assume time = iteration.
            #
            if self._ah_vars:
                # Now we find the associated times using ah.cctk_iteration
                # self.ah.cctk_iteration is a function time vs iteration, we
                # want the opposite. We define a new timeseries in which we swap
                # t and y
                self._iterations_to_times = remove_duplicated_iters(
                    self.ah.cctk_iteration.y, self.ah.cctk_iteration.t
                )
                self.shape_times = self._iterations_to_times(
                    self.shape_iterations
                )
                self.shape_time_min = self.shape_times[0]
                self.shape_time_max = self.shape_times[-1]
            else:
                warnings.warn(
                    "AH data not found, so it is impossible to convert"
                    " between iteration number to time.\nManually set"
                    " shape_times or the methods involving shape and time"
                    " will not work"
                )
                self.shape_times = None
                self.shape_time_min = None
                self.shape_time_max = None

            # We will save all the shape patches and their origin that we read in
            # this dictionary
            self._patches = {}

        # Now let's look at the VTK files
        self._vtk_files = vtk_files or {}
        self.vtk_available = self._vtk_files != {}

        self._vtk_vars: Dict[int, AttributeDictionary] = {}

    def __getitem__(self, key):
        if key not in self._qlm_vars.keys():
            raise KeyError(f"Quantity {key} does not exist")

        return self._qlm_vars[key]

    @property
    def vtk_available_iterations(self) -> List[int]:
        """Return the list of iterations at which VTK data is available.

        :returns: List of iteration with associated VTK data
        :rtype: list
        """
        return list(self._vtk_files.keys())

    def _load_vtk_at_iteration(self, iteration: int):
        """Read and load the VTK file at the given iteration.

        Raises an error when the iteration is not available.

        This operation is a little bit expensive because a long text file has to
        be parsed.

        :param iteration: Desired iteration
        :type iteration: int

        """
        if iteration not in self._vtk_files:
            raise KeyError(
                f"VTK files for iteration {iteration} not available"
            )

        # How are horizon VTK files structures?
        #
        # We learn about this in the qlm_output_vtk.F90 file in
        # QuasiLocalMeasures.
        #
        # An example looks like:
        #
        # # vtk DataFile Version 2.0
        # Horizon data
        # ASCII
        # DATASET POLYDATA
        # POINTS  2812 float
        # 1.74001573791400        3.37861181538478       0.465379287209538
        # 1.73994713406270        3.38024424250911       0.465353718016011#
        # ....
        # 1.73994920375395        3.37697921676095      -0.465402607140582
        #
        # POLYGONS  2736      13680
        # 4     0    76    77     1
        # 4     1    77    78     2
        # ...
        # 4  2734  2810  2811  2735
        # 4  2735  2811  2736  2660
        #
        # POINT_DATA  2812
        #
        # SCALARS shape float 1
        # LOOKUP_TABLE default
        # 0.465244618428368
        # 0.465266176663749
        # ...
        # 0.466705754376149
        #
        # SCALARS l0 float 1
        # LOOKUP_TABLE default
        # 2.08814072824979
        # ...
        #
        # And so on

        # So, we have:
        # 1. a first section which contains the coordinates of the points
        # 2. a second section which describes the connectivity on the mesh
        # 3. a number of sections with the values of the variables on the points

        # We have to parse all of this (we could use a library to parse it, but
        # it would be overkill)

        # To parse the file, read it in its entirety as list of strings. Then,
        # we find those indices that indicate different sections.

        with open(self._vtk_files[iteration]) as file_:
            lines = file_.readlines()

        # At which indices do the variables start? We will use this to read the
        # variables as NumPy arrays
        indices_start_variables: List[int] = []
        # Where does the coordinate section start? We'll use the POINTS word
        index_start_coordinates = None
        # Where does the connectivity section start? We'll use the POLYGON word
        index_start_connectivity = None
        for index, line in enumerate(lines):
            if line.startswith("SCALARS"):
                indices_start_variables.append(index)
            elif line.startswith("POINTS"):
                index_start_coordinates = index
            elif line.startswith("POLYGON"):
                index_start_connectivity = index

        if index_start_coordinates is None or index_start_connectivity is None:
            raise RuntimeError(
                f"VTK file {self._vtk_files[iteration]} malformed"
            )

        # We add the end of the file in indices_start_variables, so that we can
        # loop over them, except for the last one
        indices_start_variables.append(-1)

        # We extract the name of the variable by splitting on whitespces.
        variables: Dict[str, NDArray] = {}

        # Let's read the coordinates
        variables["coordinates"] = np.loadtxt(
            lines[index_start_coordinates + 1 : index_start_connectivity]
        )
        # Now the connectitivity, we have to remove 2 lines from the first occurrence of SCALARS because the
        # section start with
        #
        # POINT_DATA  2812
        #
        # This number -2 is a little fragile!
        variables["connectivity"] = np.loadtxt(
            lines[
                index_start_connectivity + 1 : indices_start_variables[0] - 2
            ],
            dtype=np.int64,
        )

        # We read the list "two elements at the time"
        for index, index_next in zip(
            indices_start_variables[:-1], indices_start_variables[1:]
        ):
            name = lines[index].split(" ")[1]
            # lines[index].split(" ") looks like: ['SCALARS', 'shape', 'float', '1\n']

            # We need index + 2 is because we have the lines like:
            #
            # SCALARS shape float 1
            # LOOKUP_TABLE default
            index_data = index + 2

            # Read the data as NumPy arrays
            variables[name] = np.loadtxt(lines[index_data:index_next])

        self._vtk_vars[iteration] = AttributeDictionary(variables)

    def available_vtk_variables_at_iteration(
        self, iteration: int
    ) -> List[str]:
        """Return the variables available in the VTK file at the given iteration.

        Most VTK variables are 1D arrays defined on the vertices of the horizon
        mesh (as in ``QuasiLocalMeasures``). The only two special VTK variables
        are ``coordiantes``, which is essentially a list of 3D coordinates of
        the vertices, and ``connectivity``, which describes how the various
        faces are formed.

        The ``connectivity`` variables is a list of indices of the form ``4 i1
        i2 i3 i4``, which means that the four vertices with indices ``i1 i2 i3
        i4`` are connected by a face. In ``QuasiLocalMeasures``, faces are
        always quadrilateral, so the first value of each element in
        ``connectivity`` will always be 4.

        :param iteration: Desired iteration
        :type iteration: int
        :returns: List of variables available
        :rtype: list

        """
        return list(self.vtk_at_iteration(iteration).keys())

    def vtk_at_iteration(self, iteration: int) -> AttributeDictionary:
        """Return a dictionary-like object containing all the VTK variables at the given iteration.

        :param iteration: Desired iteration
        :type iteration: int
        :returns: Dictionary-like object with all the VTK variables.
        :rtype: AttributeDictionary
        """
        if iteration not in self._vtk_vars:
            self._load_vtk_at_iteration(iteration)
        return self._vtk_vars[iteration]

    # Alias
    vtk_at_it = vtk_at_iteration

    def vtk_variable_at_iteration(self, key: str, iteration: int) -> NDArray:
        """Return a variable from the VTK file at the given iteration.

        See also docstring in :py:meth:`~.available_vtk_variables_at_iteration`.

        :param iteration: Desired variable
        :type iteration: int
        :param iteration: Desired iteration
        :type iteration: int
        :returns: Variable
        :rtype: NumPy array

        """
        return self.vtk_at_iteration(iteration)[key]

    def get_ah_property(self, key):
        """Return a property from AHFinderDirect as timeseries.

        :param key: AH property.
        :type key: str
        :returns: AH property as a function of time.
        :rtype: :py:class:`~.TimeSeries`
        """
        return self._ah_vars[key]

    def __str__(self) -> str:
        """Conversion to string.
        :returns: Human readable summary
        """
        ret = ""
        if self._ah_vars:
            ret += f"Formation time: {self.formation_time:.4f}\n"
        if self.shape_available:
            ret += "Shape available\n"
        if self.vtk_available:
            ret += "VTK information available\n"
        if self._qlm_vars:
            ret += f"Final Mass = {self.mass_final:.3e}\n"
            ret += f"Final Angular Momentum = {self.spin_final:.3e}\n"
            ret += f"Final Dimensionless Spin = {self.dimensionless_spin_final:.3e}"
        return ret

    def ah_origin_at_iteration(self, iteration):
        """Return the AH origin at the given iteration.

        :param iteration: Iteration number.
        :type iteration: int
        :returns: Origin of the horizon as from AHFinderDirect.
        :rtype: 3D NumPy array
        """
        return self._patches_at_iteration(iteration)[1]

    def _patches_at_iteration(self, iteration):
        """Return the shape of the horizon as a dictionary of patches with
        3D coordinates."""
        if not self.shape_available:
            raise ValueError("Shape information not available")
        if iteration not in self.shape_iterations:
            raise ValueError(
                f"Shape information for iteration {iteration} not available"
            )

        if iteration not in self._patches:
            self._patches[iteration] = self._load_patches(
                self._shape_files[iteration]
            )

        return self._patches[iteration]

    def shape_at_iteration(self, iteration):
        """Return the shape of the horizon as 3 arrays with the
        coordinates of the points.

        :param iteration: Iteration number.
        :type iteration: int
        :returns: Shape of the horizon.
        :rtype: three lists of 2D NumPy arrays, one for each
                coordinate. The list is over the different patches.
        """

        # TODO (FEATURE): Add an option to merge all the patches while keeping
        #                 an order.

        patches, _ = self._patches_at_iteration(iteration)

        # Patches is dictionary, each patch is a list of three coordiantes for
        # all the points along the two angular directions.
        coord_x = [p[0] for p in patches.values()]
        coord_y = [p[1] for p in patches.values()]
        coord_z = [p[2] for p in patches.values()]

        return coord_x, coord_y, coord_z

    def shape_at_time(self, time, tolerance=1e-10):
        """Return the shape of the horizon as 3 arrays with the
        coordinates of the points.

        :param time: Time.
        :type time: float
        :param tolerance: Tolerance in determining the time.
        :type tolerance: float
        :returns: Shape of the horizon.
        :rtype: three lists of 2D NumPy arrays, one for each
                coordinate. The list is over the different patches.
        """
        # https://stackoverflow.com/a/41022847
        try:
            index_closest = next(
                i
                for i, _ in enumerate(self.shape_times)
                if np.isclose(_, time, tolerance)
            )
        except StopIteration:
            raise ValueError(f"Time {time} not available")

        iteration = self.shape_iterations[index_closest]

        return self.shape_at_iteration(iteration)

    def shape_time_at_iteration(self, iteration):
        """Return the time corresponding to the given iteration using the information
        from the shape files.

        :param iteration: Iteration number.
        :type iteration: int
        :returns: Time at the given iteration.
        :rtype: float
        """
        if iteration not in self.shape_iterations:
            raise ValueError(f"Shape not available for iteration {iteration}")

        return self._iterations_to_times(iteration)

    @staticmethod
    def _load_patches(path):
        """AHFinderDirect uses a system of multipatches to avoid coordinates
        singularities. Each patch covers a portion around an axis (e.g., +z
        axis). We will load this data.

        """
        # TODO (FEATURE): Add support for HDF5 files

        # ASCII files are structured in this way:
        # * There is a header with the number of patches and the origin
        # * For each patch, there is a section taht starts with
        #   ### +z patch and a local header
        # * Each section has multiple groups with one angular coordinate
        #   fixed, the various groups are separated by a blank line

        # Here we match the patch name. This matches
        # 1. the entire string ^ #
        # 2. the literals ### ..... patch
        # 3. and then we find what is inside +/- and one between xyz
        rx_patch = re.compile(r"^### ([+-][xyz]) patch$")

        # Here we find the origin of the system.
        # 1. We match the entire string (^ $)
        # 2. The match the literal # origin =
        # 3. We match numbers with possibly +/- and dots, and spaces in between
        #    ([\s])
        rx_origin = re.compile(
            r"^# origin = ([+-eE\d.]+)[\s]+([+-eE\d.]+)[\s]+([+-eE\d.]+)[\s]*$"
        )
        origin = None
        with open(path, "r") as fil:
            # patches is a dictionary that maps the name of the patch to the
            # coordiantes of that patch
            patches = {}
            # We use these variables as local variables
            current_patch = None
            current_patch_data = []
            current_coordinates = []
            # We scan the entire file line by line
            for line in fil:
                # We read the header, which starts with #
                if line.startswith("#"):
                    # If we haven't found the origin yet, let's look for it
                    if origin is None:
                        matched_origin = rx_origin.match(line)
                        if matched_origin is not None:
                            origin = np.array(
                                [
                                    float(matched_origin.group(coord))
                                    for coord in (1, 2, 3)
                                ]
                            )
                            # If we know that this the origin line, we can
                            # skip the rest of the analysis
                            continue

                    matched_patch = rx_patch.match(line)
                    if matched_patch is not None:
                        # If current_patch is not None it means that we have
                        # already read an header, so here we are moving to the
                        # next patch. We save all the data to the dictionary
                        # patches, and move reset current_patch_data to an empty
                        # list.
                        if current_patch is not None:
                            patches[current_patch] = current_patch_data
                            current_patch_data = []
                        current_patch = matched_patch.group(1)
                elif (line == "") or (line.isspace()):
                    # Here we found the blank line, which means that we have
                    # to flush the data and move on to a new group in the same
                    # patch.
                    # If current_coordinates is empty, the following will test
                    # negative. If it is not empty, it means that we have read
                    # the current group, so we can reset the current_coordiantes
                    # variable
                    if current_coordinates:
                        current_patch_data.append(current_coordinates)
                        current_coordinates = []
                else:
                    # Data line should have 6 columns, the last three are x, y,
                    # and z.
                    data = line.split()
                    if len(data) != 6:
                        raise RuntimeError(f"Corrupt AH shape file {path}")
                    coordinates = list(map(float, data[3:]))
                    # coordinates is a list like:
                    # [5.198917752 -0.1549284014 0.1549284014],
                    # so we append the list
                    current_coordinates.append(coordinates)

            # Here we take care of the last group
            if current_coordinates:
                current_patch_data.append(current_coordinates)
            if current_patch is not None:
                patches[current_patch] = current_patch_data

        if origin is None:
            raise RuntimeError("Corrupt AH files, missing origin.")
        # Reorganize the data from the AHHorizonDirect format to a NumPy matrix
        # Each patch is an array with three lists, one for each direction
        patches = {
            p: np.transpose(np.array(d), axes=(2, 0, 1))
            for p, d in patches.items()
        }
        return patches, origin

    def shape_outline_at_iteration(self, iteration, cut):
        """Return the cut of the 3D shape on a specified plane.

        ``cut`` has to be a 3D tuple or list with None on the dimensions you
        want to keep, and the value of the other coordinates. For example, if
        you want the outline at z = 3 on the xy plane, ``cut`` has to be
        ``(None, None, 3)``.

        No interpolation is performed, so results are not accurate when the cut
        is not along one of the major directions centered in the origin of the
        horizon.

        :param iteration: Iteration number.
        :type iteration: int
        :param cut: How should the horizon be sliced?
        :type cut: 3D tuple
        :returns:    Coordinates of AH outline.
        :rtype:      tuple of two 1D NumPy arrays.

        """
        # TODO (FEATURE): Add interpolation for off-axis cuts
        #
        # At the moment, cuts that are not along the main axes will return
        # outlines with very few points, so we have to implement and
        # interpolation scheme to ensure that the accuracy is high.

        if iteration not in self.shape_iterations:
            raise ValueError(f"Shape not available for iteration {iteration}")

        if not isinstance(cut, (tuple, list)):
            raise TypeError(
                f"Cut has to be a list or a tuple, not a {type(cut)}"
            )

        if len(cut) != 3:
            raise ValueError("Cut has to be three-dimensional")

        patches, origin = self._patches_at_iteration(iteration)

        # If cut is three None, we return the entire shape.
        # TODO (FEATURE): Merge patches here
        #
        # We should not return the multiple patches, but a single one,
        # so we should merge them.
        if cut.count(None) == 3:  # all None
            return self.shape_at_iteration(iteration)

        # There is at least one that is not None

        # If they are all not None, that's a point, so we cannot return anything
        # meaningful
        if cut.count(None) == 0:
            raise ValueError("Cut must has some entries set to None")

        # There is at least one that is None, but not all
        #
        # This can be 1D or 2D.

        # On what dimension(s) do we have to look at?
        dims = [index for index, value in enumerate(cut) if value is None]

        # Let's first consider the 1D case
        if cut.count(None) == 1:
            # Here we just have to find those points with the specified
            # coordinates

            # Here dim is the axis we are looking at, dim0 and dim1 the other
            # two
            dim = dims[0]
            (dim0, dim1) = [index for index in range(3) if index not in dims]

            # Here we collect the result
            points = []

            # We closest points up to tolerance of 0.1%, unless the cut is
            # outside the horizon
            for patch in patches.values():
                coords, coords0, coords1 = patch[dim], patch[dim0], patch[dim1]
                # We check we are inside the horizon
                if not (
                    np.min(coords0) <= cut[dim0] <= np.max(coords0)
                    and np.min(coords1) <= cut[dim1] <= np.max(coords1)
                ):
                    continue

                size0 = np.max(coords0) - np.min(coords0)
                size1 = np.max(coords1) - np.min(coords1)
                points_around_cut0 = abs(coords0 - cut[dim0]) < 1e-3 * size0
                points_around_cut1 = abs(coords1 - cut[dim1]) < 1e-3 * size1

                selected_points = np.logical_and(
                    points_around_cut0, points_around_cut1
                )
                points.append(coords[selected_points])

            return points

        # Finally the 2D case
        #
        # The catch here is that we want to merge the patches and make sure
        # that they are in order.
        if cut.count(None) == 2:
            (dim0, dim1) = dims
            # dim is the normal direction to the plane where we want the outline
            dim = [index for index in range(3) if index not in dims][0]
            # In points0 and points1 we collect the points on the dimensions 0
            # an 1, below we also use 0 and 1 to refer to these two dimensions.
            points0, points1 = [], []

            # We closest points up to tolerance of 0.1%, unless the cut is
            # outside the horizon
            for patch in patches.values():
                coords, coords0, coords1 = patch[dim], patch[dim0], patch[dim1]
                if not (np.min(coords) <= cut[dim] <= np.max(coords)):
                    continue
                size = np.max(coords) - np.min(coords)
                points_around_cut = abs(coords - cut[dim]) < 1e-3 * size
                points0.append(coords0[points_around_cut])
                points1.append(coords1[points_around_cut])

            if len(points0) == 0:
                return None

            points0 = np.hstack(points0)
            points1 = np.hstack(points1)
            # We compute angle to properly order the points
            phi = np.angle(
                points0 - origin[dim0] + 1j * (points1 - origin[dim1])
            )
            ordering = np.argsort(phi)
            return points0[ordering], points1[ordering]

    def shape_outline_at_time(self, time, cut, tolerance=1e-10):
        """Return the cut of the 3D shape on a specified plane.

        ``cut`` has to be a 3D tuple or list with None on the dimensions you
        want to keep, and the value of the other coordinates. For example, if
        you want the outline at z = 3 on the xy plane, ``cut`` has to be
        ``(None, None, 3)``.

        No interpolation is performed, so results are not accurate when the cut
        is not along one of the major directions centered in the origin of the
        horizon.

        :param time: Time.
        :type time: float
        :param tolerance: Tolerance in determining the time.
        :type tolerance: float
        :param cut: How should the horizon be sliced?
        :type cut: 3D tuple
        :returns:    Coordinates of AH outline.
        :rtype:      tuple of two 1D NumPy arrays.

        """
        # TODO: (REFACTOR) Code logic shared with shape_at_time
        #
        # This code is essentially the same as shape_at_time. We should refactor
        # all the _at_time methods, maybe defining a function that transforms
        # _at_iteration methods to _at_time_methods.

        # https://stackoverflow.com/a/41022847
        try:
            index_closest = next(
                i
                for i, _ in enumerate(self.shape_times)
                if np.isclose(_, time, tolerance)
            )
        except StopIteration:
            raise ValueError(f"Time {time} not available")

        iteration = self.shape_iterations[index_closest]

        return self.shape_outline_at_iteration(iteration, cut=cut)


class HorizonsDir:
    """Class to collect information on apparent horizons
    available from thorns AHFinderDirect and QuasiLocalMeasures.

    :ivar found_any:   True if at least one horizon was found.

    AHFinderDirect and QuasiLocalMeasures have different indexing. You must
    provide both when accessing a file. In the future, the map between the
    two indexing systems will be inferred from the paramter file.

    """

    # What variables should not be passed to OneHorizon?
    _exclude_ah_vars = ["cctk_time"]

    def __init__(self, sd):
        """Constructor.

        :param sd:  SimDir object providing access to data directory.
        :type sd:   SimDir

        """
        # We organize the variables from QuasiLocalMeasures in the dictionary
        # _qlm_vars. This dictionary has as keys the number of qlm horizon and
        # as value another dictionary with as keys the name of the variables
        # stripped of the qlm_ prefix and of the number, and as values the
        # timeseries. We extract the number with a regular expression which
        # matches qlm_VARNAME[NUM]

        self._qlm_vars = {}
        self._populate_qlm_vars(sd)
        self._num_qlm_horizons = len(self._qlm_vars.keys())

        # self._vtk_files is a dictionary that maps QLM index to a dictionary
        # which maps iteration to full path of the corresponding vtk file
        self._vtk_files: Dict[int, Dict[int, str]] = {}
        self._populate_vtk_files(sd)

        # ah_vars is a dictionary like qlm_vars with the difference that we
        # extract information from the files BH_diagnostics.ah(\d+).gp and
        # that the index here is the Apparent Horizon index.
        self._ah_vars = {}
        # self._num_ah_horizons is set inside the function
        self._populate_ah_vars(sd)

        # The next step is to find the files for the shape of the horizons, if
        # available. We scan all the files and find those with h.t*****.ah*.gp
        #
        # Once again we put all the files in a dictionary with index the AH
        # index and as values another dictionary with keys the iteration and
        # value the file
        self._shape_files = {}
        self._populate_shape_files(sd)

        # Here we align the ah_vars and shape_files so that they have the same
        # keys. We add an empty {} to the missing values.
        for ah_index in self._ah_vars:
            if ah_index not in self._shape_files:
                self._shape_files[ah_index] = {}

        for ah_index in self._shape_files:
            if ah_index not in self._ah_vars:
                self._ah_vars[ah_index] = {}

        # Okay, here we can be confident that shape_files and ah_vars have the
        # same keys.

        # Now there are multiple options:
        # 1. We don't have any qlm_vars, ah_vars, and _shape_files
        # 2. We qlm_vars, ah_vars, and shape_files
        # 3. We have a combination of the above
        #
        # The OneHorizon class is robust in taking partially empty data,
        # but we must clear the indexing system.
        #
        # Here we check that all the dictionaries are not empty
        self.found_any = any(
            [self._qlm_vars, self._ah_vars, self._shape_files]
        )

    def _populate_qlm_vars(self, sd):
        # Here we a regular expression with two capturing groups
        # 1. ^ $ means that we match the entire string
        # 2. qlm_ is matched to itself
        # 3. the first capturing group, (\w+), matches any word
        # 4. Then we match the brackets, and inside a number
        rx_qlm_number = re.compile(r"^qlm_(\w+)\[(\d+)\]$")

        for var_name in sd.ts.scalar.keys():
            matched = rx_qlm_number.match(var_name)
            if matched is not None:
                # Here we strip of qlm_ and of the number
                var_name_stripped = matched.group(1)
                horizon_number = int(matched.group(2))
                # For each horizon, we have a dictionary that maps variable
                # names to the timeseries
                horizon_vars = self._qlm_vars.setdefault(horizon_number, {})
                horizon_vars[var_name_stripped] = sd.ts.scalar[var_name]

    def _populate_vtk_files(self, sd):
        # Here we a regular expression with two capturing groups
        # 1. ^ $ means that we match the entire string
        # 2. surface is matched to itself
        # 3. the first capturing group, (\d+), matches any number, the QLM index
        # 4. we match _
        # 5. then we the iteration number
        # 6. Finally we match .vtk
        rx_vtk_filename = re.compile(r"^surface(\d+)_(\d+).vtk$")

        for path in sd.allfiles:
            filename = os.path.split(path)[-1]
            matched = rx_vtk_filename.search(filename)
            if matched is not None:
                # We remove the leading zeros and remove 1 because
                # qlm_calculate.F90 starts counting from 1 instead of 0
                vtk_index = int(matched.group(1).lstrip("0")) - 1
                # The 'or 0' is here for the special case in which iteration is 0
                iteration = int(matched.group(2).lstrip("0") or "0")
                self._vtk_files.setdefault(vtk_index, {})[iteration] = path

    def _populate_ah_vars(self, sd):
        # First, we find all the files related to apparent horizons. These
        # have names like BH_diagnostics.ah1.gp
        self._ah_files = {}

        rx_ah_filename = re.compile(r"^BH_diagnostics.ah(\d+).gp$")
        for path in sd.allfiles:
            filename = os.path.split(path)[-1]
            matched = rx_ah_filename.search(filename)
            if matched is not None:
                ah_index = int(matched.group(1))
                self._ah_files.setdefault(ah_index, []).append(path)

        # Next, we find what variables they contain. This should be pretty
        # standard, but we can make our code more robust by not assuming too
        # much. We read one header and find the variables, then read all the
        # other files assuming the have the same variables. A complication is
        # that variables names have blank spaces. We turn them into underscores.

        self._num_ah_horizons = len(self._ah_files.keys())

        # We continue only if we find some files
        if self._num_ah_horizons > 0:
            # [0][0] is because the values are lists
            first_ah_file = tuple(self._ah_files.values())[0][0]
            with open(first_ah_file, "r") as fil:
                # Here we read the first lines_to_read into header
                # We strip the new line
                header = []
                for line in fil:
                    # We read the header, which starts with #
                    if line.startswith("#"):
                        header.append(line.strip())
                    else:
                        break

            # Now, we parse the header and associate variable name with column
            # where the data is. The header looks like:
            #
            # # apparent horizon 1/3
            # #
            # # column  1 = cctk_iteration
            # # column  2 = cctk_time
            # # column  3 = centroid_x
            # # column  4 = centroid_y
            # # column  5 = centroid_z
            #
            # We scan the columns with a regex.
            # 1. ^ $ means that we match the entire string
            # 2. \#[\s]column[\s]+ matches the literal '# column ' with any
            #    number of spaces.
            # 3. Then we match the number
            # 4. We match another literal ' = ' with the sapces
            # 5. Finally we match the name of the variable matching letters
            #    and symbols
            rx_column = re.compile(
                r"\#[\s]column[\s]+(\d+)[\s]=[\s]([a-zA-Z_0-9\s()-/]+)$"
            )

            # Here is where we store the map
            self._ah_vars_columns = {}

            for line in header:
                matched = rx_column.match(line)
                if matched is not None:
                    # Columns counting start from 1, so we must subtract one to
                    # be with 0-based indexing
                    column_number = int(matched.group(1)) - 1
                    name = matched.group(2)

                    # We need to know where the time is
                    if name == "cctk_time":
                        time_column = column_number

                    # We exclude some variables we don't want in OneHorizon
                    # (e.g., we don't want cctk_time)
                    if name in self._exclude_ah_vars:
                        continue

                    # Spaces to underscores
                    name = name.replace(" ", "_")
                    # We remove parentheses
                    name = name.replace("(", "")
                    name = name.replace(")", "")
                    # We change / to _
                    name = name.replace("/", "_")
                    # We change - to _
                    name = name.replace("-", "_")
                    self._ah_vars_columns[name] = column_number

            # Now we are ready to populate, we read all the data first. Then, we
            # select all the columns
            for ah_index, files in self._ah_files.items():
                # We create an empty dictionary in self._ah_vars[ah_index]
                self._ah_vars.setdefault(ah_index, {})

                # We read all the data
                alldata = [np.loadtxt(f, unpack=True, ndmin=2) for f in files]
                for var_name, column_number in self._ah_vars_columns.items():
                    # Here we select the time column and the data column for all
                    # the data in each file and we convert them into TimeSeries
                    data_ts = combine_ts(
                        [
                            remove_duplicated_iters(
                                data[time_column], data[column_number]
                            )
                            for data in alldata
                        ]
                    )
                    self._ah_vars[ah_index][var_name] = data_ts

    def _populate_shape_files(self, sd):
        # Here we match the files with a regular expression:
        # 1. ^ $ means that we match the entire string
        # 2. Then we match the literal h.t
        # 3. with a number (\d+)
        # 4. the literal .ah
        # 5. another number (\d+)
        # 6. and the file extension .gp
        rx_shape_filename = re.compile(r"^h.t(\d+).ah(\d+).gp$")
        for path in sd.allfiles:
            filename = os.path.split(path)[1]
            matched = rx_shape_filename.match(filename)
            if matched is not None:
                ah_index = int(matched.group(2))
                iteration = int(matched.group(1))
                ah_shape_dict = self._shape_files.setdefault(ah_index, {})
                ah_shape_dict[iteration] = path

    @property
    def available_qlm_horizons(self):
        """Horizons in QLM indexing with associated data.

        :returns: Indices of the QLM horizons found in the data.
        :rtype: list
        """
        return sorted(list(self._qlm_vars.keys()))

    @property
    def available_apparent_horizons(self):
        """Horizons in AH indexing with associated data.

        :returns: Indices of the AH horizons found in the data.
        :rtype: list"""
        return sorted(list(self._ah_vars.keys()))

    def get_apparent_horizon(self, ah):
        """Return a :py:func:`~.OneHorizon` corresponding to an horizon as
        found by AHFinderDirect.

        :param ah: AHFinderDirect index.
        :type ah: int

        :returns: Apparent horizon.
        :rtype: :py:func:`~.OneHorizon`
        """
        if ah not in self.available_apparent_horizons:
            raise KeyError(f"AH horizon {ah} not available")
        return self[(-1, ah)]

    def get_qlm_horizon(self, qlm):
        """Return a :py:func:`~.OneHorizon` corresponding to an horizon as
        analyzed by QuasiLocalMeasures.

        :param qlm: QuasiLocalMeasures index.
        :type qlm: int

        :returns: QuasiLocalMeasures horizon.
        :rtype: :py:func:`~.OneHorizon`
        """
        if qlm not in self.available_qlm_horizons:
            raise KeyError(f"QLM horizon {qlm} not available")
        return self[(qlm, -1)]

    def __getitem__(self, key):
        if not isinstance(key, (tuple, list)):
            raise TypeError("You have to provide both the QLM and AH indices")

        if len(key) != 2:
            raise KeyError(f"{key} does not identify an horizon")

        # We ensure that we have ints because we store the keys as int.
        qlm_index, ah_index = int(key[0]), int(key[1])

        # We need at least of the two indices to be valid
        if (
            qlm_index not in self.available_qlm_horizons
            and ah_index not in self.available_apparent_horizons
        ):
            raise KeyError(f"Horizon {key} is not found (QLM and AH absent)")

        # With get we return an empty dictionary if we don't have the index
        return OneHorizon(
            self._qlm_vars.get(qlm_index, {}),
            self._ah_vars.get(ah_index, {}),
            self._shape_files.get(ah_index, {}),
            self._vtk_files.get(qlm_index, {}),
        )

    def __str__(self):
        if not self.found_any:
            return "No horizon found"

        num_qlm = len(self.available_qlm_horizons)
        num_ah = len(self.available_apparent_horizons)

        ret = "Horizons found:\n"
        ret += f"{num_qlm} horizons from QuasiLocalMeasures\n"
        ret += f"{num_ah} horizons from AHFinderDirect"
        return ret

    # TODO (FUTURE): Infer the mapping from the parameter file
    #
    # Once that is done, implement the following features.

    # def __iter__(self):
    #     for horizon_number in self.available_horizons:
    #         yield self[horizon_number]

    # def __len__(self):
    #     return len(self.available_horizons)
